import logging
from threading import Timer
from typing import Dict

from socorepo import config
from socorepo.fetcher import fetch_components
from socorepo.structs import Component

log = logging.getLogger("socorepo")

component_cache: Dict[str, Dict[str, Component]] = {}


def _fetch_all_components():
    new_component_cache = {}

    for project in config.PROJECTS.values():
        try:
            new_component_cache[project.id] = fetch_components(project)
        except Exception:
            log.exception("An unexpected and uncaught exception occurred while fetching. "
                          "Will continue with the next project shortly...")

    return new_component_cache


def _periodic():
    global component_cache

    # Schedule the next iteration.
    t = Timer(config.FETCH_INTERVAL, _periodic)
    t.daemon = True
    t.start()

    # Actually update the cache.
    log.info("Will now start updating the component cache by fetching the components of all projects...")
    new_component_cache = _fetch_all_components()
    component_cache = new_component_cache
    log.info("Finished updating the component cache.")


_periodic()



import os
from itertools import dropwhile

from flask import request, url_for, send_file, render_template, jsonify, redirect, abort, make_response
from werkzeug.exceptions import HTTPException
from werkzeug.http import HTTP_STATUS_CODES

from socorepo import app, config
from socorepo.cache import component_cache
from socorepo.forms import create_component_filter_form
from socorepo.l10n import msg_lang, _, localized_of


@app.context_processor
def global_context_vars():
    return {
        "bare": config.APPEARANCE_BARE,
        "global_title": config.APPEARANCE_TITLE,
        "heading": config.APPEARANCE_HEADING,
        "footer": localized_of(config.APPEARANCE_FOOTERS),
        "msg_lang": msg_lang(),
        "_": _,
        "localized_of": localized_of
    }


@app.errorhandler(Exception)
def error(e):
    if isinstance(e, HTTPException):
        error_code = e.code
        error_desc = e.description
    else:
        error_code = 500
        error_desc = _("internal_error")
        app.log_exception(e)

    error_name = HTTP_STATUS_CODES.get(error_code, "Unknown Error")
    return render_template("error.html", error_code=error_code, error_name=error_name,
                           error_desc=error_desc), error_code


@app.route("/" + os.path.basename(config.APPEARANCE_FAVICON_PATH))
def favicon():
    return send_file(config.APPEARANCE_FAVICON_PATH)


@app.route("/")
def project_list():
    resp = render_template("project_list.html", homepage=localized_of(config.APPEARANCE_HOMEPAGES),
                           projects=config.PROJECTS.values())
    if config.APPEARANCE_BARE:
        resp = make_response(resp)
        resp.headers["Socorepo-Language"] = msg_lang()
    return resp


@app.route("/<project_id>/")
def project(project_id):
    if project_id not in config.PROJECTS:
        abort(404)

    proj = config.PROJECTS[project_id]
    components = component_cache[project_id].values()

    # Get highlight components: Latest release and latest experimental version.
    highlight_components = []
    latest_stable = next(dropwhile(lambda c: not c.qualifier.stable, components), None)
    if latest_stable:
        highlight_components.append((_("project.latest_stable"), latest_stable))
    latest_experimental = next(iter(components), None)
    if latest_experimental and not latest_experimental.qualifier.stable:
        highlight_components.append((_("project.latest_experimental"), latest_experimental))

    # Construct filter form.
    available_qualifiers = sorted(set(comp.qualifier for comp in components))
    comp_filter_form = create_component_filter_form(available_qualifiers)(request.args)
    # Apply filters, if applicable.
    if comp_filter_form.validate():
        filter_version = comp_filter_form.version.data
        filter_qualifier = comp_filter_form.qualifier.data
        components = [comp for comp in components
                      if comp.version.startswith(filter_version)
                      and (filter_qualifier == "" or comp.qualifier.name == filter_qualifier)]

    # Get list of all featured asset type matchers which have actually matched an asset...
    occurring_featured_asset_type_matchers = []
    for comp in components:
        for asset in comp.assets:
            if asset.featured and asset.matcher_causing_featuring not in occurring_featured_asset_type_matchers:
                occurring_featured_asset_type_matchers.append(asset.matcher_causing_featuring)
    # ... and sort them by the order they are referenced in the project's configuration.
    occurring_featured_asset_type_matchers.sort(key=lambda matcher: proj.featured_asset_type_matchers.index(matcher))

    resp = render_template("project.html", component_filter_form=comp_filter_form, project=proj,
                           components=components, highlight_components=highlight_components,
                           occurring_featured_asset_type_matchers=occurring_featured_asset_type_matchers)
    if config.APPEARANCE_BARE:
        resp = make_response(resp)
        resp.headers["Socorepo-Language"] = msg_lang()
        resp.headers["Socorepo-Project-Label"] = proj.label
    return resp


@app.route("/<project_id>/<version>")
def component(project_id, version):
    if project_id not in config.PROJECTS or project_id not in component_cache \
            or version not in component_cache[project_id]:
        abort(404)

    proj = config.PROJECTS[project_id]
    comp = component_cache[project_id][version]

    comp_info_tables = [loc.component_info_table(comp) for loc in proj.locators if loc.id in comp.extra_data]
    has_file_size_column = any(asset.file_size for asset in comp.assets)
    has_checksums_column = any(asset.checksums for asset in comp.assets)

    resp = render_template("component.html", project=proj, component=comp, component_info_tables=comp_info_tables,
                           has_file_size_column=has_file_size_column, has_checksums_column=has_checksums_column)
    if config.APPEARANCE_BARE:
        resp = make_response(resp)
        resp.headers["Socorepo-Language"] = msg_lang()
        resp.headers["Socorepo-Project-Label"] = proj.label
        resp.headers["Socorepo-Component-Version"] = comp.version
        resp.headers["Socorepo-Component-Qualifier"] = comp.qualifier.name
    return resp


# ===========
# === API ===
# ===========

@app.route("/api/v1/projects/")
def api_project_list():
    return jsonify({
        "projects": [{
            "id": proj.id,
            "label": proj.label
        } for proj in config.PROJECTS.values()]
    })


@app.route("/api/v1/components/<project_id>/")
def api_component_list(project_id):
    if project_id not in config.PROJECTS:
        abort(404)

    return jsonify({
        "components": [{
            "version": comp.version,
            "qualifier": comp.qualifier.name
        } for comp in component_cache[project_id].values()]
    })


@app.route("/api/v1/assets/<project_id>/<version>/")
def api_asset_list(project_id, version):
    if project_id not in config.PROJECTS or project_id not in component_cache \
            or version not in component_cache[project_id]:
        abort(404)

    return jsonify({
        "assets": [{
            "filename": asset.filename,
            "file_size": asset.file_size,
            "url": asset.url,
            "checksums": asset.checksums,
            "type": str(asset.type)
        } for asset in component_cache[project_id][version].assets]
    })


# =====================================
# === Legacy routes from QuarterMAP ===
# =====================================

@app.route("/index/")
@app.route("/projects/")
def legacy_index():
    return redirect(url_for("project_list"))


@app.route("/projects/details")
@app.route("/projects/artifactList")
def legacy_project():
    if "projectId" not in request.args:
        abort(404)
    return redirect(url_for("project", project_id=request.args["projectId"].lower()))


@app.route("/projects/artifact")
def legacy_component():
    if "projectId" not in request.args or "version" not in request.args:
        abort(404)
    return redirect(url_for("component", project_id=request.args["projectId"].lower(), version=request.args["version"]))



import logging
import os
import re
import sys
from typing import Pattern

from markupsafe import Markup

from socorepo import config
from socorepo.config.toml_dict import TomlDict
from socorepo.consts import COLORS
from socorepo.locators import LOCATOR_PARSERS
from socorepo.structs import Project, VersionQualifier, AssetClfMatcher, AssetTypeMatcher

log = logging.getLogger("socorepo")

# Determine the config dir. Make the path absolute to avoid relative path confusion.
config.EXTERNAL_CONFIG = "SOCOREPO_CONFIG_DIR" in os.environ
config_dir = os.path.abspath(os.environ["SOCOREPO_CONFIG_DIR"] if config.EXTERNAL_CONFIG
                             else os.path.join(os.path.dirname(__file__), "..", "default_config"))


def load_general_settings():
    toml_settings = TomlDict.load(os.path.join(config_dir, "settings.toml"))

    config.CONFIGURE_LOGGING = toml_settings.req("configure_logging", bool)
    config.LOG_DIR = os.path.join(config_dir, toml_settings.req("log_dir", str))
    config.APPLICATION_ROOT = toml_settings.req("application_root", str)
    config.FETCH_INTERVAL = toml_settings.req("fetch_interval", int)

    config.APPEARANCE_BARE = toml_settings.req("appearance.bare", bool)
    config.APPEARANCE_TITLE = toml_settings.req("appearance.title", str)
    config.APPEARANCE_HEADING = toml_settings.req("appearance.heading", str)
    config.APPEARANCE_FAVICON_PATH = os.path.join(config_dir, toml_settings.req("appearance.favicon", str))

    config.APPEARANCE_HOMEPAGES = _load_one_markup_file_per_lang(toml_settings.req("appearance.homepage", str))
    config.APPEARANCE_FOOTERS = _load_one_markup_file_per_lang(toml_settings.req("appearance.footer", str))


def _load_one_markup_file_per_lang(root_file):
    dir_ = os.path.dirname(os.path.join(config_dir, root_file))

    prefix = os.path.basename(root_file)
    if "." in prefix:
        prefix = prefix[:prefix.rindex(".")]

    per_lang = {}
    for filename in os.listdir(dir_):
        if filename.startswith(prefix):
            lang = filename[len(prefix):]
            if "." in lang:
                lang = lang[:lang.rindex(".")]
            lang = lang[1:]  # remove _ between prefix and language if it exists
            with open(os.path.join(dir_, filename), "r") as f:
                per_lang[lang] = Markup(f.read())
    if "" not in per_lang:
        log.error("Must supply a default '%s' file without a language suffix, used when no lang matches.", root_file)
        sys.exit()
    return per_lang


def load_remaining_config():
    _load_version_qualifiers()
    _load_asset_clfs()
    _load_projects()


def _load_version_qualifiers():
    config.VERSION_QUALIFIERS = []

    toml_qualifiers = TomlDict.load(os.path.join(config_dir, "version_qualifiers.toml"))
    for ordinal, name in enumerate(toml_qualifiers):
        toml_qualifier = toml_qualifiers.sub(name)

        if toml_qualifier.opt("default", bool, fallback=False) == ("version_section_regex" in toml_qualifier):
            log.error("%s: The 'version_section_regex' of a version qualifier must be present "
                      "if and only if default = false.", toml_qualifier.error_prefix)
            sys.exit()

        default = toml_qualifier.opt("default", bool, fallback=False)
        config.VERSION_QUALIFIERS.append(VersionQualifier(
            ordinal=ordinal,
            name=name,
            color=toml_qualifier.req("color", str, choices=COLORS),
            version_section_regex=None if default else toml_qualifier.req("version_section_regex", Pattern),
            default=default,
            stable=toml_qualifier.opt("stable", bool, fallback=False)
        ))

    # Make sure there is exactly one default version qualifier.
    defaults = [qualifier for qualifier in config.VERSION_QUALIFIERS if qualifier.default]
    if len(defaults) == 1:
        config.DEFAULT_VERSION_QUALIFIER = defaults[0]
    else:
        msg = "%s: There has to be exactly one default version classifier."
        if len(defaults) > 1:
            msg += f" You have defined {len(defaults)}: " + _quote_and_join(q.name for q in defaults)
        log.error(msg, toml_qualifiers.error_prefix)
        sys.exit()


def _load_asset_clfs():
    toml_asset_clfs = TomlDict.load(os.path.join(config_dir, "asset_classifiers.toml"))

    config.ASSET_CLF_MATCHERS = [
        # Note: Don't split the key as it might contain dots!
        AssetClfMatcher(clf=clf, filename_regex=toml_asset_clfs.req(clf, Pattern, split_key=False))
        for clf in toml_asset_clfs
    ]

    config.ASSET_CLFS = [matcher.clf for matcher in config.ASSET_CLF_MATCHERS]

    # Make sure the asset clfs do not contain any whitespace or question marks.
    clfs_with_whitespace = [clf for clf in config.ASSET_CLFS if re.search(r"[\s?]", clf)]
    if clfs_with_whitespace:
        log.error("%s: Asset classifiers are not allowed to contain any whitespace or question marks. "
                  "The following classifiers violate that rule: %s",
                  toml_asset_clfs.error_prefix, _quote_and_join(clfs_with_whitespace))
        sys.exit()


def _load_projects():
    config.PROJECTS = {}

    toml_available_templates = TomlDict.load(os.path.join(config_dir, "project_templates.toml"))
    toml_projects = TomlDict.load(os.path.join(config_dir, "projects.toml"))

    for project_id in toml_projects:
        toml_project = toml_projects.sub(project_id)

        # Recursively resolve included templates.
        toml_project.apply_templates(toml_available_templates)

        descriptions = {}
        for key in toml_project:
            if key.startswith("description"):
                lang = key[len("description_"):]
                descriptions[lang] = Markup(toml_project.req(key, str))
        if descriptions and "" not in descriptions:
            log.error("%s: Must supply a default description without a language suffix, used when no lang matches.",
                      toml_project.error_prefix)
            sys.exit()

        # Parse optional version substitutions.
        version_substitutions = []
        if "version_substitutions" in toml_project:
            for toml_version_sub in toml_project.sub_list("version_substitutions"):
                pattern = toml_version_sub.req("pattern", Pattern)
                repl = toml_version_sub.req("repl", str)
                version_substitutions.append((pattern, repl))

        excluded_asset_clfs = toml_project.opt("excluded_asset_clfs", list, fallback=[])

        try:
            featured_asset_type_matchers = [AssetTypeMatcher(pattern) for pattern
                                            in toml_project.opt("featured_asset_types", list, fallback=[])]
        except ValueError as e:
            log.error("%s: Error while parsing featured asset types: %s",
                      toml_project.error_prefix, e)
            sys.exit()

        # Make sure the referenced excluded clfs are actually defined.
        undefined_asset_clfs = set(excluded_asset_clfs).difference(config.ASSET_CLFS)
        if undefined_asset_clfs:
            log.error("%s: These asset classifiers are referenced in value behind key 'excluded_asset_clfs' "
                      "even though they have not been defined in 'asset_classifiers.toml': %s",
                      toml_project.error_prefix, _quote_and_join(undefined_asset_clfs))
            sys.exit()

        # Parse locators.
        locators = []
        for locator_id in toml_project.sub("locators"):
            toml_locator = toml_project.sub(f"locators.{locator_id}")
            locator_type = toml_locator.req("type", str, choices=LOCATOR_PARSERS)
            locator = LOCATOR_PARSERS[locator_type](locator_id, toml_locator)
            locators.append(locator)

        # Parse the rest of the project and register it to the config module.
        project = Project(
            id=project_id,
            label=toml_project.req("label", str),
            descriptions=descriptions,
            version_substitutions=version_substitutions,
            excluded_asset_clfs=excluded_asset_clfs,
            featured_asset_type_matchers=featured_asset_type_matchers,
            locators=locators
        )
        config.PROJECTS[project_id] = project


def _quote_and_join(objs):
    return ", ".join(f"'{obj}'" for obj in objs)



import string
from itertools import dropwhile
from typing import Union, List

from socorepo import config
from socorepo.structs import Component, VersionQualifier


def get_version_qualifier(version: str) -> VersionQualifier:
    for sec in _split_version(version):
        if isinstance(sec, VersionQualifier):
            return sec

    return config.DEFAULT_VERSION_QUALIFIER


def sort_components_by_version(components: List[Component]) -> List[Component]:
    return sorted(components, reverse=True,
                  key=_create_key(lambda component: _split_version(component.version),
                                  _less_than_for_split_versions))


# Splits a version into sections where either
# - a special character marks a new section or
# - a change in character class (numeric/lowercase/uppercase/other) occurs.
# Also converts numeric sections to ints.
# For example, 1.3.9-RC2b is split into [1, 3, 9, "RC", 2, "b"].
def _split_version(version: str) -> List[Union[str, int, VersionQualifier]]:
    sections = []
    start = 0
    for idx, char in enumerate(version):
        if char in string.punctuation:
            sections.append(version[start:idx])
            start = idx + 1
        elif version[start].isnumeric() != char.isnumeric():
            sections.append(version[start:idx])
            start = idx
    sections.append(version[start:])

    return list(map(_parse_version_section, sections))


def _parse_version_section(sec: str):
    if sec.isnumeric():
        return int(sec)
    else:
        for qualifier in config.VERSION_QUALIFIERS:
            if qualifier.matches(sec):
                return qualifier
        return sec


def _create_key(convert: callable, less_than: callable) -> callable:
    class Key(object):
        __slots__ = ['obj']  # RAM optimization

        def __init__(self, obj):
            self.obj = convert(obj)

        # Python guarantees to only use __lt__ when sorting.
        def __lt__(self, other):
            return less_than(self.obj, other.obj)

    return Key


# This is a slightly modified version of standard __lt__ on lists.
def _less_than_for_split_versions(split_ver_1: list, split_ver_2: list) -> bool:
    disparity = dropwhile(lambda secs: secs[0] == secs[1], zip(split_ver_1, split_ver_2))

    try:
        # --- CASE 1: The two versions differ somewhere. ---
        sec_1, sec_2 = next(disparity)

        # Special case: If the two versions differ at a section where one of the versions features a qualifier
        # and the other one doesn't, the version with the qualifier is always SMALLER (e.g., 0.3-ALPHA < 0.3.1).
        if isinstance(sec_1, VersionQualifier) and not isinstance(sec_2, VersionQualifier):
            return True
        if not isinstance(sec_1, VersionQualifier) and isinstance(sec_2, VersionQualifier):
            return False

        try:
            return sec_1 < sec_2
        except TypeError:  # in case one of the sections is a str and the other one is a number
            return str(sec_1) < str(sec_2)
    except StopIteration:
        # --- CASE 2: One version is a prefix of the other one (e.g., 0.3 and 0.3.1). ---

        # Special case: If the common prefix is immediately followed by a qualifier (e.g., 0.3 and 0.3-ALPHA),
        # the version with the qualifier is considered SMALLER, even though
        # it is longer (in contrast to standard behavior).
        if len(split_ver_1) > len(split_ver_2) and isinstance(split_ver_1[len(split_ver_2)], VersionQualifier):
            return True
        if len(split_ver_1) < len(split_ver_2) and isinstance(split_ver_2[len(split_ver_1)], VersionQualifier):
            return False

        # Standard behavior: The shorter version is considered SMALLER.
        return len(split_ver_1) < len(split_ver_2)



from flask_wtf import FlaskForm
from wtforms import StringField, SelectField, SubmitField

from socorepo import config
from socorepo.l10n import _


def create_component_filter_form(available_qualifiers):
    default_qualifier = config.DEFAULT_VERSION_QUALIFIER

    class ComponentFilterForm(FlaskForm):
        version = StringField(_("project.filter_version"))
        qualifier = SelectField(_("project.filter_qualifier"),
                                choices=[("", "Any")] + [(q.name, q.name) for q in available_qualifiers],
                                default=default_qualifier.name if default_qualifier in available_qualifiers else "")
        filter = SubmitField(_("project.filter"))

    return ComponentFilterForm



import csv
import os

from flask import request
from flask.ctx import has_request_context


def _load_translations():
    dir_ = os.path.join(os.path.dirname(__file__), "l10n")
    for filename in os.listdir(dir_):
        dct = {}
        with open(os.path.join(dir_, filename), "r") as file_handle:
            reader = csv.reader(file_handle, delimiter="=")
            for row in reader:
                if len(row) == 0:
                    continue
                elif len(row) == 2:
                    dct[row[0]] = row[1]
                else:
                    raise ValueError(f"Each non-empty row in 'l10n/{filename}' must have one non-quoted '='.")
        lang = filename[:-len(".csv")]
        _translations[lang] = dct


_translations = {}
_load_translations()


def msg_lang():
    if not has_request_context():
        raise LookupError("Can't determine message language when not currently in a request context.")
    lang = getattr(request, "lang", None)
    if lang is None:
        lang = request.accept_languages.best_match(_translations.keys(), default="en")
        request.lang = lang
    return lang


def _(key, *args, **kwargs):
    return _translations[msg_lang()][key].format(*args, **kwargs)


def localized_of(variants: dict):
    if not has_request_context():
        raise LookupError("Can't select a localized variant when we're not currently in a request context.")
    lang = request.accept_languages.best_match(variants.keys(), default="")
    return variants[lang]



from dataclasses import dataclass
from typing import Union, List, Dict
from urllib.parse import urljoin

from markupsafe import Markup

from socorepo.config.toml_dict import TomlDict
from socorepo.l10n import _
from socorepo.locators.helpers import ensure_trailing_slash
from socorepo.structs import Locator, ComponentPrototype, Component


@dataclass(frozen=True)
class Nexus3(Locator):
    server: str
    verify_tls_certificate: bool
    repository: str
    component_group: str
    component_name: str

    def fetch_component_prototypes(self) -> List[ComponentPrototype]:
        from . import fetcher  # avoid cyclic dependencies
        return fetcher.fetch_component_prototypes(self)

    def component_info_table(self, component: Component) -> Dict[str, Union[str, Markup]]:
        repository_url = urljoin(self.server, "#browse/browse:" + self.repository)

        comp_url = urljoin(self.server, "#browse/search/custom=repository_name%3D" + self.repository)
        coord_markup = ""
        if self.component_group:
            comp_url += "%20AND%20group.raw%3D" + self.component_group
            coord_markup += f'<a href="{comp_url}" target="_blank">{self.component_group}</a> &rightarrow; '
        comp_url += "%20AND%20name.raw%3D" + self.component_name
        coord_markup += f'<a href="{comp_url}" target="_blank">{self.component_name}</a> &rightarrow; '
        comp_url += "%20AND%20version%3D" + component.extra_data[self.id].nexus_version
        coord_markup += f'<a href="{comp_url}" target="_blank">{component.version}</a>'

        return {
            _("locator.source_type"): _("locator.nexus3.source_type"),
            _("locator.nexus3.repository"): Markup(
                f'<a href="{self.server}" target="_blank">{self.server}</a> &rightarrow; '
                f'<a href="{repository_url}" target="_blank">{self.repository}</a>'
            ),
            _("locator.nexus3.component_coordinates"): Markup(coord_markup)
        }


@dataclass(frozen=True)
class Nexus3ComponentData:
    nexus_version: str


def parse_locator(locator_id: str, toml_locator: TomlDict):
    return Nexus3(id=locator_id,
                  server=ensure_trailing_slash(toml_locator.req("server", str)),
                  verify_tls_certificate=toml_locator.opt("verify_tls_certificate", bool, fallback=True),
                  repository=toml_locator.req("repository", str),
                  component_group=toml_locator.opt("component_group", str),
                  component_name=toml_locator.req("component_name", str))



import logging
from collections import OrderedDict
from typing import Optional, List, Dict

from socorepo import config
from socorepo.fetcher.versions import get_version_qualifier, sort_components_by_version
from socorepo.structs import Project, AssetTypeMatcher, ComponentPrototype, AssetPrototype, Component, Asset, AssetType

log = logging.getLogger("socorepo")


def fetch_components(project: Project):
    log.info("Now fetching components for project '%s'...", project.id)
    comp_proto_list = []
    for locator in project.locators:
        log.info("Now fetching components for project '%s' using %s locator '%s'...",
                 project.id, type(locator).__name__, locator.id)
        comp_proto_list += locator.fetch_component_prototypes()
    return _new_component_dict(project, comp_proto_list)


def _new_component_dict(project: Project, comp_proto_list: List[ComponentPrototype]) -> Dict[str, Component]:
    # Group the component prototypes by version. This is necessary when multiple locators are used.
    comp_proto_groups = {}
    for comp_proto in comp_proto_list:
        version = comp_proto.version
        # Apply version substitutions if there are any.
        for pattern, replacement in project.version_substitutions:
            version = pattern.sub(replacement, version)
        comp_proto_groups.setdefault(version, []).append(comp_proto)

    # Convert the grouped component prototypes to actual components.
    comp_list = [_new_component(project, version, comp_proto_group)
                 for version, comp_proto_group in comp_proto_groups.items()]

    # Sort components in order of decreasing version.
    sorted_comp_list = sort_components_by_version(comp_list)

    # Transform the list of components into a dict which maps from a version to a component.
    return OrderedDict([(comp.version, comp) for comp in sorted_comp_list])


def _new_component(project: Project, version: str, comp_proto_group: List[ComponentPrototype]) -> Component:
    # Collect all asset prototypes of all component prototypes in the group.
    asset_protos = [asset for comp_proto in comp_proto_group for asset in comp_proto.assets]

    # Determine the type (i.e., a list of asset clfs) of each asset prototype.
    asset_types = {asset_proto: _determine_asset_type(project, asset_proto)
                   for asset_proto in asset_protos}

    # Sort the asset prototypes by (1) clfs as ordered in the config file and (2) filename.
    sorted_asset_protos = sorted(asset_protos,
                                 key=lambda asset_proto: (
                                     [config.ASSET_CLFS.index(clf) for clf in asset_types[asset_proto].clfs],
                                     asset_proto.filename)
                                 )

    # For each asset prototype, find the featured asset type matcher which matches it.
    # Note that each matcher can only match one asset and is taken out of the game afterwards.
    # Given that, it is important to know that the assets are matched in sorted order.
    #
    # Example: If a matcher matches "Py*", only the asset with the highest python version will be matched by it,
    #          since the "Py<version>" clfs are defined in order of decreasing version
    #          (at least in the default asset_classifiers.toml).
    asset_matchers = {}
    remaining_matchers = project.featured_asset_type_matchers.copy()
    for asset_proto in sorted_asset_protos:
        matcher = _find_matching_matcher(project, remaining_matchers, version, asset_proto, asset_types[asset_proto])
        asset_matchers[asset_proto] = matcher
        if matcher is not None:
            remaining_matchers.remove(matcher)

    # Convert the asset prototypes to actual assets using the obtained information.
    sorted_assets = [_new_asset(asset_proto, asset_types[asset_proto], asset_matchers[asset_proto])
                     for asset_proto in sorted_asset_protos]

    return Component(version=version,
                     qualifier=get_version_qualifier(version),
                     assets=sorted_assets,
                     extra_data={comp_proto.locator_id: comp_proto.extra_data for comp_proto in comp_proto_group})


def _determine_asset_type(project: Project, asset_proto: AssetPrototype) -> AssetType:
    # Note that this selection preserves the order of the clfs as they are defined in the config file.
    return AssetType([matcher.clf for matcher in config.ASSET_CLF_MATCHERS
                      if (matcher.matches(asset_proto.filename) or matcher.clf in asset_proto.forced_clfs)
                      and matcher.clf not in project.excluded_asset_clfs])


def _find_matching_matcher(project: Project, matchers: List[AssetTypeMatcher], version: str,
                           asset_proto: AssetPrototype, type_: AssetType) -> Optional[AssetTypeMatcher]:
    matching_matchers = [m for m in matchers if m.matches(type_)]
    if len(matching_matchers) > 1:
        log.warning("In project '%s', component with version '%s': "
                    "Asset with filename '%s' matches multiple featured asset types: %s. "
                    "This is discouraged. All but the first match will now be ignored. "
                    "Try to rewrite your featured asset types such that each asset only matches one of them.",
                    project.id, version, asset_proto.filename,
                    ", ".join(f"'{m.pattern}'" for m in matching_matchers))
    return next(iter(matching_matchers), None)


def _new_asset(asset_proto: AssetPrototype, type_: AssetType, matcher: AssetTypeMatcher) -> Asset:
    return Asset(filename=asset_proto.filename,
                 file_size=asset_proto.file_size,
                 url=asset_proto.url,
                 checksums=OrderedDict(sorted(asset_proto.checksums.items())),
                 type=type_,
                 featured=matcher is not None,
                 matcher_causing_featuring=matcher)



import logging

import urllib3
from flask import Flask
from werkzeug.middleware.dispatcher import DispatcherMiddleware

from socorepo import config
from socorepo.config.loader import load_general_settings, load_remaining_config
from socorepo.log import setup_logging


def app_root_404(env, resp):
    resp("404", [("Content-Type", "text/plain")])
    return [b"404 The application root has been reconfigured."]


__version__ = "1.2.0"

# Disable unverified TLS certificate warnings.
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

# Load the basic config, which we require for setting up logging.
load_general_settings()

if config.CONFIGURE_LOGGING:
    setup_logging(config.LOG_DIR if config.EXTERNAL_CONFIG else None)

# Now, we can load the remaining config. Any errors here will be logged to the logfile.
load_remaining_config()

if not config.EXTERNAL_CONFIG:
    logging.getLogger("socorepo").warning(
        "Running off internally stored default configuration files. This might not be what you want. "
        "See README for more information on how to use your own configuration.")

# Create the app.
app = Flask(__name__)
app.config["WTF_CSRF_ENABLED"] = False

# Change the application root if configured.
if config.APPLICATION_ROOT != "/":
    app.config["APPLICATION_ROOT"] = config.APPLICATION_ROOT
    app.wsgi_app = DispatcherMiddleware(app_root_404, {config.APPLICATION_ROOT: app.wsgi_app})

# Initialize routes.
from . import views

# Start the cache scheduler.
from . import cache
